/*
 * zc_str.c
 * Author: SAITOH Akira (1999,2001)
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif/*HAVE_CONFIG_H*/

#define _GNU_SOURCE 1
#include <math.h>
#ifndef NAN
#ifdef NaN
#define NAN NaN
#else
#define NAN 0.0
#endif/*NaN*/
#endif/*NAN*/
#include <stdlib.h>
#include <string.h>
#include "zcomplex.h"
#include "zc_str.h"
#define N_MAX 1000
#define PI 3.141592653589793238462643383279502884197169\
3993751058209749445923078164062862
/*Return true when a string s1 has s2 in itself.*/
static int s1_has_s2 (char *s1, const char *s2)
{
   int m, n, ns1, ns2;
   ns1 = strlen (s1);
   ns2 = strlen (s2);
   for (m = 0; m < ns1 - ns2 + 1; m++)
      if (*(s1 + m) == *s2)
      {
         for (n = 1; n < ns2; n++)
            if (*(s1 + m + n) != *(s2 + n))
               break;
         if (n == ns2)
            return 1;
      }
   return 0;
}

/*Searching from position p, return true when a string s1 has s2 in itself.*/
/******
static int s1_has_s2_p (char *s1, const char *s2, int p)
{
   int m, n, ns1, ns2;
   ns1 = strlen (s1);
   ns2 = strlen (s2);
   for (m = p; m < ns1 - ns2 + 1; m++)
      if (*(s1 + m) == *s2)
      {
         for (n = 1; n < ns2; n++)
            if (*(s1 + m + n) != *(s2 + n))
               break;
         if (n == ns2)
            return 1;
      }
   return 0;
}
******/

/*Trim spaces in str. str must be terminated with '\0'.*/
static void remove_spaces (char *str)
{
   int m, n, N = N_MAX;
   char *buf = NULL;
   if (N > (n = strlen (str)))
      N = n;
   buf = (char *) malloc ((N + 1) * sizeof(char));
   if (buf == NULL) return;
   for (m = n = 0; n < N; n++)
      if (*(str + n) != ' ')
      {
         *(buf + m) = *(str + n);
         m++;
      }
   *(buf + m) = '\0';/*We can do this because m =< N.*/
   strcpy (str, buf);
   free (buf);
}

/*Check if the string represents a complex number.*/
int is_zc_str (char *str)
{
   int n, N = N_MAX;
   if (N > (n = strlen (str)))
      N = n;
   for (n = 0; n < N; n++)
      switch ((int) *(str + n))
      {
         case '0': case '1': case '2': case '3': case '4': case '5':
         case '6': case '7': case '8': case '9': case '.': case 'e':
         case 'E': case '+': case '-': case 'x': case 'h': case 'X':
         case 'H': case ' ': break;

         case 'i':
         case 'j': return 1;

         default: return 0;
      }
   return 0;
}

/*
 * For a debug use only:
 */
/*#include <stdio.h>
 *#define DBG_R {printf("Line %d: buf0 is %s, buf1 is %s\n", __LINE__, buf0, buf1);}
 */
/*
 * When we do not debug, use this:
 */
#define DBG_R 


/*
 *Create a zcomplex number from some string.
 *"a2zc" means "alphabet to zcomplex".
 */
zcomplex a2zc (char *str)
{
   int n, N = N_MAX;
   int was_iorj = 0, was_plus = 0;
   int nb0 = 0, nb1 = 0;
   zcomplex w = {NAN, NAN, NAN, NAN};
   char *buf0, *buf1, *b_str;

   /*Keep str as it is.*/
   b_str = (char *) malloc ((strlen (str) + 1) * sizeof(char));
   if (b_str == NULL) return (w);

   strcpy (b_str, str);
   remove_spaces (b_str);
   /*By removing spaces, we may obtain following terms or something.
    *a: ___+j___ or ___-j___
    *b: ___+___j or ___-___j
    *c: ___+ej___pi or ___-ej_pi__
    */
   if (N > (n = strlen (b_str)))
      N = n;

   buf0 = (char *) malloc ((N + 1) * sizeof(char));
   if (buf0 == NULL)
   {
      free (b_str);
      return (w);
   }
   buf1 = (char *) malloc ((N + 1) * sizeof(char));
   if (buf1 == NULL)
   {
      free (b_str);
      free (buf0);
      return (w);
   }

   for (n = 0; n < N; n++)
      switch ((int) *(b_str + n))
      {
         case 'i':
         case 'j':
            if (was_plus)
            {
            /*___+___i*/
               if ((*(b_str + n - 1) != '+' && *(b_str + n - 1) != '-') ||
                   (n - 2 < 0 ? 0 : (*(b_str + n - 2) == 'e' || *(b_str + n - 2) == 'E')))
               {
                  *(buf0 + nb0) = *(buf1 + nb1) = '\0';
                  w.r = atof (buf0); w.i = atof (buf1);
                  absarg (&w);
                  DBG_R
                  free (buf0);
                  free (buf1);
                  free (b_str);
                  return (w);
               }
            /*___+i*/
               else if (n == N - 1)
               {
                  *(buf0 + nb0) = '\0';
                  w.r = atof (buf0);
                  if (*(b_str + n - 1) == '-') w.i = -1.0;
                  else w.i = 1.0;
                  absarg (&w);
                  DBG_R
                  free (buf0);
                  free (buf1);
                  free (b_str);
                  return (w);
               }
            /*___+i___*/
               else
               {
                  was_iorj = 1;
                  DBG_R
                  break;
               }
            }
          /*___i or +i___*/
            else if (n != 0)
            {
               if (n == 1 && N > 2)
               {
                  /*+i___*/
                  int l;
                  for (nb0 = 0, l = 2; l < N; nb0++, l++)
                     *(buf0 + nb0) = *(b_str + l);
                  *(buf0 + nb0) = '\0';
                  w.r = 0.0; w.i = atof (buf0);
                  if (*b_str == '-')
                     w.i *= -1.0;
                  absarg (&w);
                  DBG_R
                  free (buf0);
                  free (buf1);
                  free (b_str);
                  return (w);
               }
               /*___i*/
               *(buf0 + nb0) = '\0';
               w.r = 0.0; w.i = atof (buf0);
               if (n == 1)
               {
                  /*+i*/
                  if (*b_str == '+')
                     w.i = 1.0;
                  /*-i*/
                  else if (*b_str == '-')
                     w.i = -1.0;
               }
               absarg (&w);
               DBG_R
               free (buf0);
               free (buf1);
               free (b_str);
               return (w);
            }
            else
            {
            /*i*/
               if (n == N - 1)
               {
                  w.r = 0.0; w.i = 1.0;
                  absarg (&w);
                  DBG_R
                  free (buf0);
                  free (buf1);                  
                  free (b_str);
                  return (w);
               }
            /*i___*/
               else
               {
                  was_iorj = 1;
                  break;
               }
            }
         
         case '+':
         case '-':
         /*___+___i*/
            if (n != 0)
               if (*(b_str + n - 1) != 'e' && *(b_str + n - 1) != 'E')
                  was_plus = 1;

         case 'e':
            if (*(b_str + n) == 'e')
            if (n < N - 1)
            if (*(b_str + n + 1) == 'i' || *(b_str + n + 1) == 'j')
            {
               int l, m;
               for (m = 0; m < n + 1; m++)
                  *(buf0 + m) = *(b_str + m);
               for (l = 0, m++; m < N; m++, l++)
                  *(buf1 + l) = *(b_str + m);
               *(buf0 + n + 1) = *(buf1 + l) = '\0';
               w.abs = atof (buf0); w.arg = atof (buf1);
               /*___ej or ___ejpi*/
               if (n + 2 == N ||
                   *(b_str + n + 2) == 'p' || *(b_str + n + 2) == 'P')
                  w.arg = 1.0;
               /*___ej__pi or ej__pi etc*/
               if (s1_has_s2 (buf1, "pi") || s1_has_s2 (buf1, "PI"))
                  w.arg *= PI;
               /*___ej__deg or ej__deg etc*/
               if (s1_has_s2 (buf1, "deg") || s1_has_s2 (buf1, "DEG"))
                  w.arg *= PI/180.0;
               /*ej___ or ej___pi etc*/
               if (n == 0) w.abs = 1.0;
               if (w.abs < 0.0)
               {
                  w.abs *= -1.0;
                  w.arg += (w.arg > 0.0 ? -PI : PI);
               }
               realimag (&w);
               DBG_R
               free (buf0);
               free (buf1);
               free (b_str);
               return (w);
            }

         default:
            if (was_iorj || was_plus)
            {
               *(buf1 + nb1) = *(b_str + n);
               if (*(buf1 + nb1) != '+') nb1++;
            }
            else
            {
               *(buf0 + nb0) = *(b_str + n);
               nb0++;
            }
            break;
      }
   *(buf0 + nb0) = *(buf1 + nb1) = '\0';
   w.r = atof (buf0); w.i = atof (buf1);
   absarg (&w);
   DBG_R
   free (buf0);
   free (buf1);
   free (b_str);
   return (w);
}
