/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 *Functions for complex number calculation
 * Author: Akira Saitoh <s-akira at users.sourceforge.net> (1997-2005)
 * I think algorithms in this file may be not so attractive, but
 * I have been spending time on improving the accuracy of calculation.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif/*HAVE_CONFIG_H*/

#define _GNU_SOURCE 1
#include <math.h>
#ifndef NAN
#ifdef NaN
#define NAN NaN
#else
#define NAN (0.0/0.0)
#endif/*NaN*/
#endif/*NAN*/

#ifndef HAVE_ISINF
#ifdef isinf
#undef isinf
#endif /*isinf*/
#ifdef HAVE_IEEEFP_H
#include <ieeefp.h>
#define isinf(x) ((!finite((x)))*((x)<0.0?-1:1))
#else
#define isinf is_inf
static int is_inf (double x)
{
   if (x == HUGE_VAL) return (1);
   if (x == -HUGE_VAL) return (-1);
   return (0);
}
#endif /*HAVE_IEEEFP_H*/
#endif /*HAVE_ISINF*/

#include "zcomplex.h"

/*0.5PI*/
#define sPI    1.5707963267948966192313216916397514420985\
846996875529104874722961539082031431
/*PI*/
#define PI     3.1415926535897932384626433832795028841971\
693993751058209749445923078164062862
/*2PI*/
#define dPI    6.2831853071795864769252867665590057683943\
387987502116419498891846156328125724
/*1/(2PI)*/
#define RdPI   1.5915494309189533576888376337251436203445\
964574045644874766734405889679763423e-1
/*exp(10)*/
#define expTen 2.2026465794806716516957900645284244366353\
512618556781074235426355225202818571e4
/*
 *The argument value is between -PI and PI by default.
 *If the domain [0, 2PI) is needed, bt0_2PI should be 1.
 */
int bt0_2PI = 0;
/* But I know Japanese students prefer the argument between 0 and 2PI.*/
/*int bt0_2PI = 1;*/
/*
 *The maximum for the value 'n' of Hermite Polynomials.
 */
/*
int max_n_of_hermite = 300;
*/

/*round off values experimentally*/
static zcomplex roundoff_0 (zcomplex z);
/*a bit more reliable functions*/
static double SIN (double x);
static double COS (double x);
static double EXP (double x);
static double SINH (double x);
static double COSH (double x);
/*Ordinary mod function*/
static double r_mod (double a, double b);

static zcomplex roundoff_0 (zcomplex z)
{
  zcomplex c;
  double d;
  
  c = z;
  d = z.abs * 3.333e-16;
  if (d < 1.0e-5)
  {
     if (-d < c.r && c.r < d) c.r = 0.0;
     if (-d < c.i && c.i < d) c.i = 0.0;
  }
  absarg(&c);
  return (c);
}

static double SIN(double x)
{
  double y;
  y = x;
  
  if(x > dPI)
         y -= dPI * floor(x*RdPI);
  else if(x < -dPI)
         y += dPI * floor(x*RdPI);
  return(sin(y));
}

static double COS(double x)
{
  double y;
  y = x;
  
  if(x > dPI)
         y -= dPI * floor(x*RdPI);
  else if(x < -dPI)
         y += dPI * floor(x*RdPI);
  return(cos(y));
}

static double EXP (double x)
{
   double y;
#ifndef FOR_MINGW32   
   if (x < 0.6 && x > -0.6)
      y = expm1(x) + 1.0;
   else
      y = exp(x);
#else
   /*In Mingw32, expm1 does not exist.*/
      y = exp(x);
#endif /*FOR_MINGW32*/

   return (y);
}

static double SINH (double x)
{
   double y;
   
   y = (EXP(x) - EXP(-x)) / 2.0;
   return (y);
}

static double COSH (double x)
{
   double y;
   
   y = (EXP(x) + EXP(-x)) / 2.0;
   return (y);
}

/*calculate a.abs and a.arg from a.r and a.i*/
void absarg(zcomplex *a)
{
  a->abs = hypot(a->r, a->i);
  if (a->abs == 0.0)
    a->arg = 0.0;
  else
    a->arg = atan2(a->i, a->r);
  if (a->arg == -PI) a->arg = PI;
  if (a->i < 0.0 && bt0_2PI)
    a->arg += dPI;
}
/*calculate a.r and a.i from a.abs and a.arg*/
void realimag(zcomplex *a)
{
  double d;
  
  d = a->abs * 3.333e-16;
  a->r = a->abs * COS(a->arg);
  a->i = a->abs * SIN(a->arg);
  if (d < 1.0e-5)
  {
    if (-d < a->r && a->r < d) a->r = 0.0;
    if (-d < a->i && a->i < d) a->i = 0.0;  
  }
}
/*Create a complex number from the real part and the imaginary part.*/
/*
zcomplex zcreate(double real, double imag)
{
  zcomplex c = {real,imag,0,0};
  absarg(&c);
  return(c);
}
*/
/*Create a complex number from the absolute value and the argument.*/
/*
zcomplex zcreate2(double abs, double arg)
{
  zcomplex c = {0,0,abs,arg};
  realimag(&c);
  return(c);
}
*/
/**/
zcomplex zconj(zcomplex a)/*complex conjugate*/
{
  zcomplex c;
  
  c.r = a.r; c.i = -a.i;
  absarg(&c);
  return(c);
}

zcomplex zcm(zcomplex a)/*minus*/
{
  zcomplex c;
  
  c.r = -a.r;
  c.i = -a.i;
  absarg(&c);
  return(c);
}

zcomplex zcadd(zcomplex a, zcomplex b)
{
  zcomplex c;
  
  c.r = a.r + b.r;
  c.i = a.i + b.i;
  absarg(&c);
  return(c);
}

zcomplex zcsub(zcomplex a, zcomplex b)
{
  zcomplex c;
  
  c.r = a.r - b.r;
  c.i = a.i - b.i;
  absarg(&c);
  return(c);
}

zcomplex zcmul(zcomplex a, zcomplex b)
{
  zcomplex c;
  
  c.r = a.r * b.r - a.i * b.i;
  c.i = a.r * b.i + a.i * b.r;
  absarg(&c);
  return(c);
}

zcomplex zcdiv(zcomplex a, zcomplex b)
{
  zcomplex c;
  long double d;/*denominator*/
  
  if (b.abs == 0.0 && (a.r != 0.0 && a.i == 0.0))
  {
    c.r = a.r / b.r;
    c.i = a.i;
  }
  else
  {
    d = (long double)( b.r * b.r + b.i * b.i);
    c.r = (double)((long double)(a.r * b.r + a.i * b.i) / d);
    c.i = (double)((long double)(a.i * b.r - a.r * b.i) / d);
  }
  absarg(&c);
  return(c);
}

zcomplex zcpow(zcomplex a, zcomplex b)
{ 
  zcomplex c,d,e;
  
  c = roundoff_0 (a);
  d = roundoff_0 (b);
  if (b.abs == 0.0)
  {
    zcomplex z = {1.0, 0.0, 1.0, 0.0};
    return (z);
  }
  else if ((c.abs == 0.0 && c.arg == 0.0) && d.arg == 0.0)
  {
    zcomplex z = {0.0, 0.0, 0.0, 0.0};
    return (z);
  }
  else if (b.i == 0.0)
  {
    c.abs = pow(a.abs, b.r);
    c.arg = fmod((a.arg * b.r), dPI);
    if (c.arg > PI)
      c.arg -= dPI;
    if (c.arg < -PI)
      c.arg += dPI;
    realimag(&c);
    if (c.i < 0.0 && bt0_2PI)
      c.arg += dPI;
    return (c);
  }
  c = zclog(a);
  d = zcmul(b,c);
  e = zcexp(d);
  return(e);
}
/*Ordinary mod function*/
static double r_mod (double a, double b)
{
   double r = 0.0;
   double c = 0.0;
   if (b == 0.0) return (NAN);
   /*This function must return a non-negative value.*/
   if (b < 0.0) b *= -1;
   r = a / b;
   c = ceil (r);
   if (c < r + 1.0e-12) r = c; 
   r = a - b * floor(r);
   return r;
}
/*complex fmod function*/
zcomplex zcfmod(zcomplex a, zcomplex b)
{
   double p = 0.0;
   double q = 0.0;
   double rpq = 0.0;
   zcomplex f = {0.0, 0.0, 0.0, 0.0};
   zcomplex m = {NAN, NAN, NAN, NAN};
   
   absarg (&a); absarg (&b);
   if (b.abs == 0.0) return (m);
   if (a.abs == 0.0) return (f);
   if ((fabs(a.r) > a.abs - a.abs * 1.0e-12) &&
       (fabs(b.r) > b.abs - b.abs * 1.0e-12))
   {
      a.i = b.i = 0.0;
      absarg(&a);
      absarg(&b);
   }
   if ((float)a.arg != (float)b.arg &&
       (float)a.arg != (float)(b.arg + PI) &&
       (float)a.arg != (float)(b.arg - PI))
      return (m);
   p = (a.r < 0.0 ? -a.abs : a.abs);
   q = (b.r < 0.0 ? -b.abs : b.abs);
   if (a.r == 0.0) p = (a.i < 0.0 ? -a.abs : a.abs);
   if (b.r == 0.0) q = (b.i < 0.0 ? -b.abs : b.abs);
   rpq = r_mod (p, q);
   m.abs = rpq;
   m.arg = (b.r > 0.0 ? b.arg : b.arg - PI);
   if (b.r == 0.0) m.arg = sPI;
   if (m.abs == 0.0) m.arg = 0.0;
   realimag(&m);
   if (bt0_2PI)
   {
      if (m.arg < 0.0) m.arg += dPI;
   }
   else
   {
      if (m.arg + PI < 0.0) m.arg += dPI;
   }
   return (m);
}
/**/
zcomplex zclog(zcomplex a)
{
  zcomplex c = {0.0, 0.0, 0.0, 0.0};
  
  if (a.r == 1.0 && a.i == 0.0) return(c);
  else if (a.abs < 5.0) c.r = log((a.abs * expTen)) - 10.0;
  else c.r = log(a.abs);
  c.i = a.arg;
  absarg(&c);
  return(c);
}

zcomplex zcexp(zcomplex a)
{
  zcomplex c;
#ifndef FOR_MINGW32  
  if(a.abs < 0.6)
  {
    c.r = (expm1(a.r) + 1.0) * COS(a.i);
    c.i = (expm1(a.r) + 1.0) * SIN(a.i);
  }
  else
  {
    c.r = exp(a.r) * COS(a.i);
    c.i = exp(a.r) * SIN(a.i);
  }
#else
/*In Mingw32, expm1 does not exist.*/
    c.r = exp(a.r) * COS(a.i);
    c.i = exp(a.r) * SIN(a.i);
#endif /*FOR_MINGW32*/
  absarg(&c);
  return(c);
}
/**/
zcomplex zcsin(zcomplex a)
{
  zcomplex c;
  
  c.r = SIN(a.r) * COSH(a.i);
  c.i = COS(a.r) * SINH(a.i);
  if (a.r == 0.0) c.r = 0.0;
  absarg(&c);
  return(c);
}

zcomplex zccos(zcomplex a)
{
  zcomplex c;
  
  c.r = COS(a.r) * COSH(a.i);
  c.i = -SIN(a.r) * SINH(a.i);
  if (a.r == 0.0) c.i = 0.0;
  absarg(&c);
  return(c);
}

zcomplex zctan(zcomplex a)
{
  zcomplex d;
  
  d = zcdiv(zcsin(a), zccos(a));
  return(d);
}

zcomplex zcsinh(zcomplex a)
{
  zcomplex c = {0.0, -1.0, 1.0, -sPI};
  zcomplex d = {0.0, 1.0, 1.0, sPI};

  if (a.i == 0.0)
  {
    zcomplex b = {0.0, 0.0, 0.0, 0.0};
    b.r = SINH(a.r);
    absarg(&b);
    return b;
  }
  return(zcmul(c, zcsin(zcmul(d, a))));
}

zcomplex zccosh(zcomplex a)
{
  zcomplex c = {0.0, 1.0, 1.0, sPI};
  
  if (a.i == 0.0)
  {
    zcomplex b = {0.0, 0.0, 0.0, 0.0};
    b.r = COSH(a.r);
    absarg(&b);
    return b;
  }
  return(zccos(zcmul(c, a)));
}

zcomplex zctanh(zcomplex a)
{
  if ((isinf(a.r) == 1 || a.r > 700.0) &&
      (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex b = {1.0, 0.0, 1.0, 0.0};
    return b;
  }
  if ((isinf(a.r) == -1 || a.r < -700.0) &&
      (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex b = {-1.0, 0.0, 1.0, PI};
    return b;
  }
  return(zcdiv(zcsinh(a), zccosh(a)));
}

zcomplex zcasin(zcomplex a)
{
  zcomplex b = {1.0, 0.0, 1.0, 0.0};
  zcomplex c = {0.5, 0.0, 0.5, 0.0};
  zcomplex d = {2.0, 0.0, 2.0, 0.0};
  zcomplex e = {0.0, 1.0, 1.0, sPI};
  zcomplex me = {0.0, -1.0, 1.0, -sPI};
  zcomplex f,g;
  
  if (isinf(a.r) == 1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex h = {sPI, 0.0, sPI, 0.0};
    h.i = -a.r;
    absarg(&h);
    return h;
  }
  if (isinf(a.r) == -1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex h = {-sPI, 0.0, 0.0, sPI};/*Do not modify this initial value.*/
    h.i = -a.r;
    h.abs = -a.r;
    return h;
  }
  if (a.r > 6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {sPI, 0.0, sPI, 0.0};
	h.i = -log (2.0 * a.r);
	absarg(&h);
	return h;
  }
  if (a.r < -6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {-sPI, 0.0, sPI, PI};
	h.i = log (2.0 * fabs(a.r));
	absarg(&h);
	return h;
  }
  if (a.i > 6.0e7 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
	zcomplex h = {0.0, 0.0, 0.0, 0.0};
	h.i = log (2.0 * a.i);
	absarg(&h);
	return h;
  }
  if (a.i < -6.0e7 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
	zcomplex h = {0.0, 0.0, 0.0, 0.0};
	h.i = -log (2.0 * fabs(a.i));
	absarg(&h);
	return h;
  }
  f = zcpow(zcsub(b, zcpow(a, d)), c);
  g = zcmul(me, zclog(zcadd(zcmul(e, a), f)));
  return(roundoff_0(g));
}

zcomplex zcacos(zcomplex a)
{
  zcomplex b = {1.0, 0.0, 1.0, 0.0};
  zcomplex c = {0.5, 0.0, 0.5, 0.0};
  zcomplex d = {2.0, 0.0, 2.0, 0.0};
  zcomplex e = {0.0, 1.0, 1.0, sPI};
  zcomplex me = {0.0, -1.0, 1.0, -sPI};
  zcomplex f,g;
  
  if (isinf(a.r) == 1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex h = {0.0, 0.0, 0.0, sPI};/*Do not modify this initial value.*/
    h.i = h.abs = a.r;
    return h;
  }
  if (isinf(a.r) == -1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    zcomplex h = {PI, 0.0, PI, -sPI};
    h.i = a.r;
    absarg(&h);
    return h;
  }
  if (a.r > 6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {0.0, 0.0, 0.0, 0.0};
	h.i = log (2.0 * a.r);
	absarg(&h);
	return h;
  }
  if (a.r < -6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {PI, 0.0, PI, 0.0};
	h.i = -log (2.0 * fabs(a.r));
	absarg(&h);
	return h;
  }
  if (a.i > 6.0e7 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
	zcomplex h = {sPI, 0.0, sPI, 0.0};
	h.i = -log (2.0 * a.i);
	absarg(&h);
	return h;
  }
  if (a.i < -6.0e7 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
	zcomplex h = {sPI, 0.0, sPI, 0.0};
	h.i = log (2.0 * fabs(a.i));
	absarg(&h);
	return h;
  }

  
  f = zcpow(zcsub(b, zcpow(a, d)), c);
  g = zcmul(me, zclog(zcadd(a, zcmul(e, f))));
  return(roundoff_0(g));
}

zcomplex zcatan(zcomplex a)
{
  zcomplex c = {0.0, 0.5, 0.5, sPI};
  zcomplex d = {0.0, 1.0, 1.0, sPI};
  zcomplex e = {sPI, 0.0, sPI, 0.0};

  if (isinf(a.r) == 1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
  }
  else if (isinf(a.r) == -1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    e = zcm(e);
  }
  else if (isinf(a.i) == 1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    e = zcm(e);
  }
  else if (isinf(a.i) == -1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
  }
  else
  {
    e = zcmul(c, zclog(zcdiv(zcadd(d, a), zcsub(d, a))));
  }
  return(roundoff_0(e));
}

zcomplex zcasinh(zcomplex a)
{
  zcomplex c = {2.0, 0.0, 2.0, 0.0};
  zcomplex d = {1.0, 0.0, 1.0, 0.0};
  zcomplex e = {0.5, 0.0, 0.5, 0.0};
  zcomplex f,g;
  
  if (isinf(a.r) && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
  	a.i = 0.0;
  	absarg(&a);
  	return a;
  }
  if (isinf(a.i) == 1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    zcomplex h = {0.0, sPI, sPI, sPI};
    h.r = a.i;
    absarg(&h);
    return h;
  }
  if (isinf(a.i) == -1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    zcomplex h = {0.0, -sPI, sPI, -sPI};
    h.r = -a.i;
    absarg(&h);
    return h;
  }
  if (a.r < -6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {0.0, 0.0, 0.0, 0.0};
	h.r = -log (2.0 * fabs(a.r));
	absarg(&h);
	return h;
  }
  f = zcadd(zcpow(a, c), d);
  g = zclog(zcadd(a, zcpow(f, e)));
  return(roundoff_0(g));
}

zcomplex zcacosh(zcomplex a)
{
  zcomplex c = {2.0, 0.0, 2.0, 0.0};
  zcomplex d = {-1.0, 0.0, -1.0, PI};
  zcomplex e = {0.5, 0.0, 0.5, 0.0};
  zcomplex f,g;
  
  if (isinf(a.r) == 1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
  	a.i = 0.0;
  	absarg(&a);
  	return a;
  }
  if (isinf(a.r) == -1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
  	a.i = PI;
  	absarg(&a);
  	return a;
  }
  if (isinf(a.i) == 1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    zcomplex h = {0.0, sPI, sPI, sPI};
    h.r = a.i;
    absarg(&h);
    return h;
  }
  if (isinf(a.i) == -1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    zcomplex h = {0.0, sPI, sPI, sPI};
    h.r = a.i;
    absarg(&h);
    return h;
  }
  if (a.r < -6.0e7 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
	zcomplex h = {0.0, PI, PI, sPI};
	h.r = -log (2.0 * fabs(a.r));
	absarg(&h);
	return h;
  }
  f = zcadd(zcpow(a, c), d);
  g = zclog(zcadd(a, zcpow(f, e)));
  return(roundoff_0(g));
}

zcomplex zcatanh(zcomplex a)
{
  zcomplex c = {0.5, 0.0, 0.5, 0.0};
  zcomplex d = {1.0, 0.0, 1.0, 0.0};
  zcomplex e = {0.0, -sPI, sPI, -sPI};
  if (isinf(a.r) == 1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    absarg(&e);
  }
  else if (isinf(a.r) == -1 && (a.i < 1.0e-12 && a.i > -1.0e-12))
  {
    e = zcm(e);
  }
  else if (isinf(a.i) == 1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    e = zcm(e);
  }
  else if (isinf(a.i) == -1 && (a.r < 1.0e-12 && a.r > -1.0e-12))
  {
    absarg(&e);
  }
  else
  {
    e = zcmul(c, zclog(zcdiv(zcadd(d, a), zcsub(d, a))));
  }
  return(roundoff_0(e));
}

/*Calculate gamma(a) by the asymptotic series.*/
/*coefficients*/
#define sqrt2PI 2.50662827463100050241576528481\
1045253006986740609938316629923576342293654607841
#define C1 0.0833333333333333333333333333333333\
3333333333333333333333333333333333333333333333
#define C2 0.0034722222222222222222222222222222\
22222222222222222222222222222222222222222222222
#define C3 -0.002681327160493827160493827160493\
827160493827160493827160493827160493827160493827
#define C4 -0.000229472093621399176954732510288\
0658436213991769547325102880658436213991769547325
#define C5 0.0007840392217200666274740348814422\
888496962571036645110719184793258867332941407015
#define C6 0.0000697281375836585777429398828575\
7833082935963594399808391577938903041783700219914
#define C7 -0.000592166437353693882864836225604\
4011873915851967978168252516675013245657964450831
/**/
zcomplex zcgamma(zcomplex a)
{
  int too_small = 0;
  int has_negative_real = 0;/*Whether Re[a] < 0 or not*/
  zcomplex s = {sqrt2PI,0.0,sqrt2PI,0.0}, v = {-0.5,0.0,0.5,PI};
  zcomplex w = {0.0,0.0,0.0,0.0}, W = {1.0,0.0,1.0,0.0};
  zcomplex z, ans;

  z = roundoff_0 (a);
#ifndef FOR_MINGW32  
  if (z.i == 0.0)
  {
    extern int signgam;
    double g;
    g = lgamma(a.r);
    g = (double)signgam * EXP(g);
    s.r = g;
    absarg(&s);
    return (s);
  }
#else
  /*In Mingw32, there is no signgam.*/
  if (z.i == 0.0 && (a.r < 0.0 || a.r == 0.0))
  {
    double f,c;
    f = floor(a.r);
    c = ceil(a.r);
    if (a.r == f || a.r == c)
    {
      s.r = HUGE_VAL;
      absarg(&s);
      return (s);
    }
  }
#endif /*FOR_MINGW32*/

  if (a.r < 0.0)
  {
    has_negative_real = 1;
    a = zcm(a);
  }

  z = a;

  if (a.abs < 5.0)
  {
    too_small = 1;
    z.r += 10.0; absarg(&z);
  }

  s = zcmul(s, zcexp(zcm(z)));
  s = zcmul(s, zcpow(z, zcadd(z, v)));

  w.r = C1; absarg(&w);
  v = z;
  W = zcadd(W, zcdiv(w, v));

  w.r = C2; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  w.r = C3; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  w.r = C4; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  w.r = C5; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  w.r = C6; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  w.r = C7; absarg(&w);
  v = zcmul(v, z);
  W = zcadd(W, zcdiv(w, v));

  ans = zcmul(s, W);

  if (too_small)
  {
    int i = 1;
    s = v = a;
    for ( ; i < 10; i++)
    {
       v.r += 1.0; absarg(&v);
       s = zcmul(s, v);
    }
    ans = zcdiv(ans, s);
  }

  if (has_negative_real)
  {
    zcomplex w1 = {PI,0.0,PI,0.0};
    zcomplex w2 = {-PI,0.0,PI,PI};
    w1 = zcsin(zcmul(w1, a));
    w1 = zcmul(a, zcmul(ans, w1));
    ans = zcdiv(w2, w1);
  }

  return (ans);
}
/**/
/*Hermite polynomial*/
/*"max_n_of_hermite" must be well considered.*/
/*[1].Hn'(z)=nHn-1(z), weight w(x)=exp[(-x^2)/2]*/
/*
zcomplex Hn(int n, zcomplex a)
{  
   int i;
   zcomplex zcnan = {NAN, NAN, NAN, NAN};
   zcomplex c = {0.0, 0.0, 0.0, 0.0};
   zcomplex f0 = {1.0, 0.0, 1.0, 0.0};
   zcomplex f1;
   zcomplex fnm2 = f0;
   zcomplex fnm1;
   zcomplex fn;
   
   if (n < 0 || n > max_n_of_hermite) return (zcnan);
   else if (n == 0) return (f0);
   else if (n == 1) return (a);
   fnm1 = f1 = zcm(a);
   for (i = 2; i < n + 1; i++)
   {
      c.r = (double)(1 - i);
      fn = zcadd(zcmul(f1, fnm1), zcmul(c, fnm2));
      fnm2 = fnm1; fnm1 = fn;
   }
   if (n % 2 == 1) fn = zcm(fn);
   return (fn);
}
*/
/*[2].Hn'(z)=2nHn-1(z), weight w(x)=exp[-x^2]*/
/*"max_n_of_hermite" must be considered.*/
/*
zcomplex Hn_s(int n, zcomplex a)
{  
   int i;
   zcomplex zcnan = {NAN, NAN, NAN, NAN};
   zcomplex c = {0.0, 0.0, 0.0, 0.0};
   zcomplex f0 = {1.0, 0.0, 1.0, 0.0};
   zcomplex f1;
   zcomplex fnm2 = f0;
   zcomplex fnm1;
   zcomplex fn;
   
   if (n < 0 || n > max_n_of_hermite) return (zcnan);
   else if (n == 0) return (f0);
   else if (n == 1) return (zcadd(a, a));
   fnm1 = f1 = zcm(zcadd(a, a));
   for (i = 2; i < n + 1; i++)
   {
      c.r = (double)(2 - 2 * i);
      fn = zcadd(zcmul(f1, fnm1), zcmul(c, fnm2));
      fnm2 = fnm1; fnm1 = fn;
   }
   if (n % 2 == 1) fn = zcm(fn);
   return (fn);
}
*/
